@@ -18,6 +18,7 @@
 # Refer to the README and COPYING files for full details of the license
 #
 
+import collections
 import os
 import errno
 import logging
@@ -415,41 +416,62 @@ class FileStorageDomain(sd.StorageDomain):
         """
         Return dict {volUUID: ((imgUUIDs,), parentUUID)} of the domain.
 
-        Template self image is the 1st term in template volume entry images.
-        The parent can't be determined in file domain without reading the
-        metadata.
-        Setting parent = None for compatibility with block version.
+        (imgUUIDs,) is a tuple of all the images that contain a certain
+        volUUID.  For non-templates volumes, this tuple consists of a single
+        image.  For template volume it consists of all the images that are
+        based on the template volume. In that case, the first imgUUID in the
+        tuple is the self-image of the template.
+
+        The parent of a non-template volume cannot be determined in file domain
+        without reading  the metadata. However, in order to have an output
+        compatible to block domain, we report parent as None.
+
+        Template volumes have no parent, and thus we report BLANK_UUID as their
+        parentUUID.
         """
         volMetaPattern = os.path.join(self.mountpoint, self.sdUUID,
                                       sd.DOMAIN_IMAGES, "*", "*.meta")
         volMetaPaths = self.oop.glob.glob(volMetaPattern)
-        volumes = {}
+
+        # First create mapping from images to volumes
+        images = collections.defaultdict(list)
         for metaPath in volMetaPaths:
             head, tail = os.path.split(metaPath)
             volUUID, volExt = os.path.splitext(tail)
             imgUUID = os.path.basename(head)
-            if volUUID in volumes:
-                # Templates have no parents
-                volumes[volUUID]['parent'] = sd.BLANK_UUID
-                # Template volumes are hard linked in every image directory
-                # which is derived from that template, therefore:
-                # 1. a template volume which is in use will appear at least
-                # twice (in the template image dir and in the derived image
-                # dir)
-                # 2. Any volume which appears more than once in the dir tree is
-                # by definition a template volume.
-                # 3. Any image which has more than 1 volume is not a template
-                # image. Therefore if imgUUID appears in more than one path
-                # then it is not a template.
-                if len(tuple(vPath for vPath in volMetaPaths
-                             if imgUUID in vPath)) > 1:
-                    # Add template additonal image
-                    volumes[volUUID]['imgs'].append(imgUUID)
+            images[imgUUID].append(volUUID)
+
+        # Using images to volumes mapping, we can create volumes to images
+        # mapping, detecting template volumes and template images, based on
+        # these rules:
+        #
+        # Template volumes are hard linked in every image directory
+        # which is derived from that template, therefore:
+        #
+        # 1. A template volume which is in use will appear at least twice
+        #    (in the template image dir and in the derived image dir)
+        #
+        # 2. Any volume which appears more than once in the dir tree is
+        #    by definition a template volume.
+        #
+        # 3. Any image which has more than 1 volume is not a template
+        #    image.
+
+        volumes = {}
+        for imgUUID, volUUIDs in images.iteritems():
+            for volUUID in volUUIDs:
+                if volUUID in volumes:
+                    # This must be a template volume (rule 2)
+                    volumes[volUUID]['parent'] = sd.BLANK_UUID
+                    if len(volUUIDs) > 1:
+                        # This image is not a template (rule 3)
+                        volumes[volUUID]['imgs'].append(imgUUID)
+                    else:
+                        # This image is a template (rule 3)
+                        volumes[volUUID]['imgs'].insert(0, imgUUID)
                 else:
-                    # Insert at head the template self image
-                    volumes[volUUID]['imgs'].insert(0, imgUUID)
-            else:
-                volumes[volUUID] = {'imgs': [imgUUID], 'parent': None}
+                    volumes[volUUID] = {'imgs': [imgUUID], 'parent': None}
+
         return dict((k, sd.ImgsPar(tuple(v['imgs']), v['parent']))
                     for k, v in volumes.iteritems())
 
