@@ -27,11 +27,11 @@ import glob
 import logging
 import re
 import errno
-import time
 from collections import namedtuple
 
 import misc
 from vdsm.config import config
+from vdsm.netinfo import getRouteDeviceTo
 import devicemapper
 from threading import RLock
 
@@ -96,6 +96,31 @@ def getIscsiConnectionPath(sessionId):
                         "connection%d:0" % sessionId)
 
 
+def getIscsiHostPath(sessionID):
+    """
+    Report the iSCSI host path of the given iSCSI session to be further
+    used to retrieve iface.net_ifacename (netdev) in use for the session.
+    Arguments:
+        sessionID - the iSCSI session ID.
+    Returns:
+        - iSCSI host path - e.g. '/sys/class/iscsi_host/host17/'
+    """
+    path = os.path.realpath(getIscsiSessionPath(sessionID))
+    if not os.path.exists(path):
+        raise OSError(errno.ENOENT, "No such session %r" % sessionID)
+
+    # Session path depends on the iSCSI implementation:
+    # Hardware iSCSI:
+    #   /sys/devices/pci*/*/*/host0/session7/iscsi_session/session7
+    # Software iSCSI:
+    #   /sys/devices/platform/host5/session8/iscsi_session/session8
+    host = path.rsplit(os.sep, 4)[-4]
+    if not host.startswith('host'):
+        raise RuntimeError("Unexpected session path %r" % path)
+
+    return '/sys/class/iscsi_host/' + host
+
+
 def readSessionInfo(sessionID):
     iscsi_session = getIscsiSessionPath(sessionID)
     iscsi_connection = getIscsiConnectionPath(sessionID)
@@ -113,15 +138,19 @@ def readSessionInfo(sessionID):
     paddr = os.path.join(iscsi_connection, "persistent_address")
     pport = os.path.join(iscsi_connection, "persistent_port")
 
+    # iscsi_host is available only when the session exists.
+    iscsi_host = getIscsiHostPath(sessionID)
+    netdev = os.path.join(iscsi_host, "netdev")
+
     res = []
-    for fname in (targetname, tpgt, user, passwd, paddr, pport, iface):
+    for fname in (targetname, iface, tpgt, user, passwd, paddr, pport, netdev):
         try:
             with open(fname, "r") as f:
                 res.append(f.read().strip())
         except (OSError, IOError):
             res.append("")
 
-    iqn, tpgt, username, password, ip, port, iface = res
+    iqn, iface, tpgt, username, password, ip, port, netdev = res
     port = int(port)
     tpgt = int(tpgt)
 
@@ -132,7 +161,10 @@ def readSessionInfo(sessionID):
     if password in ["<NULL>", "(null)"]:
         password = None
 
-    iface = IscsiInterface(iface)
+    if netdev in ["<NULL>", "(null)"]:
+        netdev = None
+
+    iface = IscsiInterface(iface, netIfaceName=netdev)
     portal = IscsiPortal(ip, port)
     target = IscsiTarget(portal, tpgt, iqn)
     cred = None
@@ -161,6 +193,9 @@ def addIscsiNode(iface, target, credentials=None):
                     iscsiadm.node_update(iface.name, portalStr, target.iqn,
                                          key, value, hideValue=True)
 
+            setRpFilterIfNeeded(iface.netIfaceName, target.portal.hostname,
+                                True)
+
             iscsiadm.node_login(iface.name, portalStr, target.iqn)
 
             iscsiadm.node_update(iface.name, portalStr, target.iqn,
@@ -183,6 +218,7 @@ def removeIscsiNode(iface, target):
             pass
 
         iscsiadm.node_delete(iface.name, portalStr, target.iqn)
+        setRpFilterIfNeeded(iface.netIfaceName, target.portal.hostname, False)
 
 
 def addIscsiPortal(iface, portal, credentials=None):
@@ -277,7 +313,8 @@ class IscsiInterface(object):
         'transport': ("iface.transport_name", 'r'),
         'hardwareAddress': ("iface.hwaddress", 'rw'),
         'ipAddress': ('iface.ipaddress', 'rw'),
-        'initiatorName': ('iface.initiatorname', 'rw')
+        'initiatorName': ('iface.initiatorname', 'rw'),
+        'netIfaceName': ('iface.net_ifacename', 'rw')
     }
 
     def __getattr__(self, name):
@@ -318,10 +355,11 @@ class IscsiInterface(object):
         self._conf[key] = value
 
     def __init__(self, name, hardwareAddress=None, ipAddress=None,
-                 initiatorName=None):
+                 initiatorName=None, netIfaceName=None):
 
         # Only new tcp interfaces are supported for now
         self._conf = {'iface.transport_name': 'tcp'}
+        self._conf['iface.net_ifacename'] = netIfaceName
 
         self.name = name
         self._loaded = False
@@ -357,7 +395,7 @@ class IscsiInterface(object):
         # If this fails mid operation we get a partially updated interface.
         # Suggestions are welcome.
         for key, value in self._conf.iteritems():
-            if key == 'iface.iscsi_ifacename':
+            if value is None or key == 'iface.iscsi_ifacename':
                 continue
 
             iscsiadm.iface_update(self.name, key, value)
@@ -371,40 +409,21 @@ class IscsiInterface(object):
         self._conf = conf
 
     def __repr__(self):
-        return "<IscsiInterface name='%s' transport='%s'>" % (self.name,
-                                                              self.transport)
+        return "<IscsiInterface name='%s' transport='%s' netIfaceName='%s'>" \
+            % (self.name, self.transport, self.netIfaceName)
 
 
 def iterateIscsiInterfaces():
-    names = iscsiadm.iface_list()
-    for name in names:
-        yield IscsiInterface(name)
+    for iface in iscsiadm.iface_list():
+        yield IscsiInterface(iface.ifacename, netIfaceName=iface.net_ifacename)
 
 
 @misc.samplingmethod
-def rescan(minTimeout=None, maxTimeout=None):
-    # FIXME: This whole thing is wrong from the core. We need to make rescan
-    #        completely async and have methods timeout on their own if they
-    #        can't find the devices they are looking for
-    if minTimeout is None:
-        minTimeout = config.getint('irs', 'scsi_rescan_minimal_timeout')
-    if maxTimeout is None:
-        maxTimeout = config.getint('irs', 'scsi_rescan_maximal_timeout')
-
-    if (minTimeout > maxTimeout or minTimeout < 0):
-        minTimeout = 2
-        maxTimeout = 30
-        log.warning("One of the following configuration arguments has an "
-                    "illegal value: scsi_rescan_minimal_timeout or "
-                    "scsi_rescan_maximal_timeout. Set to %s and %s seconds "
-                    "respectively.", minTimeout, maxTimeout)
-
-    log.debug("Performing SCSI scan, this will take up to %s seconds",
-              maxTimeout)
-
+def rescan():
+    timeout = config.getint('irs', 'scsi_rescan_maximal_timeout')
+    log.debug("Performing SCSI scan, this will take up to %s seconds", timeout)
     rescanOp = iscsiadm.session_rescan_async()
-    time.sleep(minTimeout)
-    rescanOp.wait(timeout=(maxTimeout - minTimeout))
+    rescanOp.wait(timeout=timeout)
 
 
 def devIsiSCSI(dev):
@@ -499,9 +518,53 @@ def disconnectFromUndelyingStorage(devPath):
 def disconnectiScsiSession(sessionID):
     # FIXME : Should throw exception on error
     sessionID = int(sessionID)
+    sessionInfo = getSessionInfo(sessionID)
     try:
         iscsiadm.session_logout(sessionID)
     except iscsiadm.IscsiError as e:
         return e[0]
 
+    netIfaceName = sessionInfo.iface.netIfaceName
+    hostname = sessionInfo.target.portal.hostname
+    setRpFilterIfNeeded(netIfaceName, hostname, False)
+
     return 0
+
+
+def _sessionsUsingNetiface(netIfaceName):
+    """ Return sessions using netIfaceName """
+    for session in iterateIscsiSessions():
+        if session.iface.netIfaceName == netIfaceName:
+            yield session
+
+
+def setRpFilterIfNeeded(netIfaceName, hostname, loose_mode):
+    """
+    Set rp_filter to loose or strict mode if there's no session using the
+    netIfaceName device and it's not the device used by the OS to reach the
+    'hostname'.
+    loose mode is needed to allow multiple iSCSI connections in a multiple NIC
+    per subnet configuration. strict mode is needed to avoid the security
+    breach where an untrusted VM can DoS the host by sending it packets with
+    spoofed random sources.
+
+    Arguments:
+        netIfaceName: the device used by the iSCSI session
+        target: iSCSI target object cointaining the portal hostname
+        loose_mode: boolean
+    """
+    if netIfaceName is None:
+        log.info("iSCSI iface.net_ifacename not provided. Skipping.")
+        return
+
+    sessions = _sessionsUsingNetiface(netIfaceName)
+
+    if not any(sessions) and netIfaceName != getRouteDeviceTo(hostname):
+        if loose_mode:
+            log.info("Setting loose mode rp_filter for device %r." %
+                     netIfaceName)
+            supervdsm.getProxy().set_rp_filter_loose(netIfaceName)
+        else:
+            log.info("Setting strict mode rp_filter for device %r." %
+                     netIfaceName)
+            supervdsm.getProxy().set_rp_filter_strict(netIfaceName)
